@@ -10,7 +10,7 @@
  *             Damien Lespiau <damien.lespiau@intel.com>
  *
  * Copyright (C) 2006 OpenedHand
- * Copyright (C) 2010 Intel Corporation
+ * Copyright (C) 2010, 2011 Intel Corporation
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -45,13 +45,38 @@
 #include <gio/gio.h>
 #include <gst/gst.h>
 #include <gst/video/video.h>
+#include <gst/interfaces/streamvolume.h>
 
 #include "clutter-gst-debug.h"
-#include "clutter-gst-private.h"
 #include "clutter-gst-enum-types.h"
+#include "clutter-gst-marshal.h"
+#include "clutter-gst-private.h"
 #include "clutter-gst-video-sink.h"
 #include "clutter-gst-video-texture.h"
 
+/* Elements don't expose header files */
+typedef enum {
+  GST_PLAY_FLAG_VIDEO         = (1 << 0),
+  GST_PLAY_FLAG_AUDIO         = (1 << 1),
+  GST_PLAY_FLAG_TEXT          = (1 << 2),
+  GST_PLAY_FLAG_VIS           = (1 << 3),
+  GST_PLAY_FLAG_SOFT_VOLUME   = (1 << 4),
+  GST_PLAY_FLAG_NATIVE_AUDIO  = (1 << 5),
+  GST_PLAY_FLAG_NATIVE_VIDEO  = (1 << 6),
+  GST_PLAY_FLAG_DOWNLOAD      = (1 << 7),
+  GST_PLAY_FLAG_BUFFERING     = (1 << 8),
+  GST_PLAY_FLAG_DEINTERLACE   = (1 << 9)
+} GstPlayFlags;
+
+enum
+{
+  DOWNLOAD_BUFFERING,
+
+  LAST_SIGNAL
+};
+
+static guint signals[LAST_SIGNAL] = { 0, };
+
 struct _ClutterGstVideoTexturePrivate
 {
   GstElement *pipeline;
@@ -61,11 +86,23 @@ struct _ClutterGstVideoTexturePrivate
   guint can_seek : 1;
   guint in_seek : 1;
   guint is_idle : 1;
+  guint is_changing_uri : 1;
+  guint in_error : 1;
+  guint in_eos : 1;
+  guint in_download_buffering : 1;
+
+  /* when in progressive download, we use the buffer-fill property to signal
+   * that we have enough data to play the stream. This flag allows to send
+   * the notify that buffer-fill is 1.0 only once */
+  guint virtual_stream_buffer_signalled : 1;
+
   gdouble stacked_progress;
+
   gdouble target_progress;
-  GstState stacked_state;
+  GstState target_state;
 
   guint tick_timeout_id;
+  guint buffering_timeout_id;
 
   /* width / height (in pixels) of the frame data before applying the pixel
    * aspect ratio */
@@ -79,6 +116,9 @@ struct _ClutterGstVideoTexturePrivate
   guint texture_width;
   guint texture_height;
 
+  /* This is a cubic volume, suitable for use in a UI cf. StreamVolume doc */
+  gdouble volume;
+
   gdouble buffer_fill;
   gdouble duration;
   gchar *font_name;
@@ -88,6 +128,8 @@ struct _ClutterGstVideoTexturePrivate
   CoglColor idle_color_unpre;
 
   GstSeekFlags seek_flags;    /* flags for the seek in set_progress(); */
+
+  GstElement *download_buffering_element;
 };
 
 enum {
@@ -109,8 +151,9 @@ enum {
   PROP_SEEK_FLAGS
 };
 
-
-#define TICK_TIMEOUT 0.5
+/* idle timeouts (in ms) */
+#define TICK_TIMEOUT        500
+#define BUFFERING_TIMEOUT   250
 
 static void clutter_media_init (ClutterMediaIface *iface);
 
@@ -120,6 +163,158 @@ G_DEFINE_TYPE_WITH_CODE (ClutterGstVideoTexture,
                          G_IMPLEMENT_INTERFACE (CLUTTER_TYPE_MEDIA,
                                                 clutter_media_init));
 
+static const gchar *
+gst_state_to_string (GstState state)
+{
+  switch (state)
+    {
+    case GST_STATE_VOID_PENDING:
+      return "pending";
+    case GST_STATE_NULL:
+      return "null";
+    case GST_STATE_READY:
+      return "ready";
+    case GST_STATE_PAUSED:
+      return "paused";
+    case GST_STATE_PLAYING:
+      return "playing";
+    }
+
+  return "Unknown state";
+}
+
+static void set_subtitle_uri (ClutterGstVideoTexture *video_texture,
+                              const gchar            *uri);
+static void configure_buffering_timeout (ClutterGstVideoTexture *video_texture,
+                                         guint                   ms);
+
+static void
+clear_download_buffering (ClutterGstVideoTexture *video_texture)
+{
+  ClutterGstVideoTexturePrivate *priv = video_texture->priv;
+
+  if (priv->download_buffering_element)
+    {
+      g_object_unref (priv->download_buffering_element);
+      priv->download_buffering_element = NULL;
+    }
+  configure_buffering_timeout (video_texture, 0);
+  priv->in_download_buffering = FALSE;
+  priv->virtual_stream_buffer_signalled = 0;
+}
+
+static gboolean
+buffering_timeout (gpointer data)
+{
+  ClutterGstVideoTexture *video_texture = CLUTTER_GST_VIDEO_TEXTURE (data);
+  ClutterGstVideoTexturePrivate *priv = video_texture->priv;
+  gdouble start_d, stop_d, seconds_buffered;
+  gint64 start, stop, left;
+  GstState current_state;
+  GstElement *element;
+  GstQuery *query;
+  gboolean res;
+
+  element = priv->download_buffering_element;
+  if (element == NULL)
+    element = priv->pipeline;
+
+  /* queue2 only knows about _PERCENT and _BYTES */
+  query = gst_query_new_buffering (GST_FORMAT_PERCENT);
+  res = gst_element_query (element, query);
+
+  if (res == FALSE)
+    {
+      priv->buffering_timeout_id = 0;
+      clear_download_buffering (video_texture);
+      return FALSE;
+    }
+
+  /* signal the current range */
+  gst_query_parse_buffering_stats (query, NULL, NULL, NULL, &left);
+  gst_query_parse_buffering_range (query, NULL, &start, &stop, NULL);
+
+  CLUTTER_GST_NOTE (BUFFERING,
+                    "start %" G_GINT64_FORMAT ", stop %" G_GINT64_FORMAT
+                    ", buffering left %" G_GINT64_FORMAT, start, stop, left);
+
+  start_d = (gdouble) start / GST_FORMAT_PERCENT_MAX;
+  stop_d = (gdouble) stop / GST_FORMAT_PERCENT_MAX;
+
+  g_signal_emit (video_texture,
+                 signals[DOWNLOAD_BUFFERING], 0, start_d, stop_d);
+
+  /* handle the "virtual stream buffer" and the associated pipeline state.
+   * We pause the pipeline until 2s of content is buffered. With the current
+   * implementation of queue2, start is always 0, so even when we seek in
+   * the stream the start position of the download-buffering signal is
+   * always 0.0. FIXME: look at gst_query_parse_nth_buffering_range () */
+  seconds_buffered = priv->duration * (stop_d - start_d);
+  priv->buffer_fill = seconds_buffered / 2.0;
+  priv->buffer_fill = CLAMP (priv->buffer_fill, 0.0, 1.0);
+
+  if (priv->buffer_fill != 1.0 || !priv->virtual_stream_buffer_signalled)
+    {
+      CLUTTER_GST_NOTE (BUFFERING, "buffer holds %0.2fs of data, buffer-fill "
+                        "is %.02f", seconds_buffered, priv->buffer_fill);
+
+      g_object_notify (G_OBJECT (video_texture), "buffer-fill");
+
+      if (priv->buffer_fill == 1.0)
+        priv->virtual_stream_buffer_signalled = 1;
+    }
+
+  gst_element_get_state (priv->pipeline, &current_state, NULL, 0);
+  if (priv->buffer_fill < 1.0)
+    {
+      if (current_state != GST_STATE_PAUSED)
+        {
+          CLUTTER_GST_NOTE (BUFFERING, "pausing the pipeline");
+          gst_element_set_state (priv->pipeline, GST_STATE_PAUSED);
+        }
+    }
+  else
+    {
+      if (current_state != priv->target_state)
+        {
+          CLUTTER_GST_NOTE (BUFFERING, "restoring the pipeline");
+          gst_element_set_state (priv->pipeline, priv->target_state);
+        }
+    }
+
+  /* the file has finished downloading */
+  if (left == G_GINT64_CONSTANT (0))
+    {
+      priv->buffering_timeout_id = 0;
+
+      clear_download_buffering (video_texture);
+      gst_query_unref (query);
+      return FALSE;
+    }
+
+  gst_query_unref (query);
+  return TRUE;
+}
+
+static void
+configure_buffering_timeout (ClutterGstVideoTexture *video_texture,
+                             guint                   ms)
+{
+  ClutterGstVideoTexturePrivate *priv = video_texture->priv;
+
+  if (priv->buffering_timeout_id)
+    {
+      g_source_remove (priv->buffering_timeout_id);
+      priv->buffering_timeout_id = 0;
+    }
+
+  if (ms)
+    {
+      priv->buffering_timeout_id =
+        g_timeout_add (ms, buffering_timeout, video_texture);
+    }
+}
+
 /* Clutter 1.4 has this symbol, we don't want to depend on 1.4 just for that
  * just yet */
 static void
@@ -335,42 +530,66 @@ set_uri (ClutterGstVideoTexture *video_texture,
   GObject *self = G_OBJECT (video_texture);
   GstState state, pending;
 
+  CLUTTER_GST_NOTE (MEDIA, "setting uri %s", uri);
+
   if (!priv->pipeline)
     return;
 
   g_free (priv->uri);
 
-  if (uri) 
+  priv->in_eos = FALSE;
+  priv->in_error = FALSE;
+
+  if (uri)
     {
       priv->uri = g_strdup (uri);
 
       /* Ensure the tick timeout is installed.
-       * 
+       *
        * We also have it installed in PAUSED state, because
        * seeks etc may have a delayed effect on the position.
        */
       if (priv->tick_timeout_id == 0)
         {
           priv->tick_timeout_id =
-            g_timeout_add (TICK_TIMEOUT * 1000, tick_timeout, self);
+            g_timeout_add (TICK_TIMEOUT, tick_timeout, self);
         }
 
       /* try to load subtitles based on the uri of the file */
+      set_subtitle_uri (video_texture, NULL);
       autoload_subtitle (video_texture, uri);
+
+      /* reset the states of download buffering */
+      clear_download_buffering (video_texture);
     }
-  else 
+  else
     {
       priv->uri = NULL;
 
-      if (priv->tick_timeout_id != 0)
+      if (priv->tick_timeout_id)
 	{
 	  g_source_remove (priv->tick_timeout_id);
 	  priv->tick_timeout_id = 0;
 	}
+
+      if (priv->buffering_timeout_id)
+        {
+          g_source_remove (priv->buffering_timeout_id);
+          priv->buffering_timeout_id = 0;
+        }
+
+      if (priv->download_buffering_element)
+        {
+          g_object_unref (priv->download_buffering_element);
+          priv->download_buffering_element = NULL;
+        }
+
     }
 
   priv->can_seek = FALSE;
   priv->duration = 0.0;
+  priv->stacked_progress = 0.0;
+  priv->target_progress = 0.0;
 
   CLUTTER_GST_NOTE (MEDIA, "setting URI: %s", uri);
 
@@ -380,11 +599,13 @@ set_uri (ClutterGstVideoTexture *video_texture,
       if (pending)
         state = pending;
 
-      gst_element_set_state (priv->pipeline, GST_STATE_NULL);
+      gst_element_set_state (priv->pipeline, GST_STATE_READY);
 
       g_object_set (priv->pipeline, "uri", uri, NULL);
 
       gst_element_set_state (priv->pipeline, state);
+
+      priv->is_changing_uri = TRUE;
     }
   else
     {
@@ -414,23 +635,24 @@ set_playing (ClutterGstVideoTexture *video_texture,
 
   CLUTTER_GST_NOTE (MEDIA, "set playing: %d", playing);
 
-  if (priv->uri) 
+  priv->in_error = FALSE;
+  priv->in_eos = FALSE;
+
+  priv->target_state = playing ? GST_STATE_PLAYING : GST_STATE_PAUSED;
+
+  if (priv->uri)
     {
-      GstState state = GST_STATE_PAUSED;
-
-      if (playing)
-	state = GST_STATE_PLAYING;
-
       priv->in_seek = FALSE;
 
-      gst_element_set_state (priv->pipeline, state);
-    } 
-  else 
+      g_message ("set to %d", priv->target_state);
+      gst_element_set_state (priv->pipeline, priv->target_state);
+    }
+  else
     {
       if (playing)
-	g_warning ("Unable to start playing: no URI is set");
+       g_warning ("Unable to start playing: no URI is set");
     }
-  
+
   g_object_notify (G_OBJECT (video_texture), "playing");
   g_object_notify (G_OBJECT (video_texture), "progress");
 }
@@ -444,9 +666,9 @@ get_playing (ClutterGstVideoTexture *video_texture)
 
   if (!priv->pipeline)
     return FALSE;
-  
+
   gst_element_get_state (priv->pipeline, &state, &pending, 0);
-  
+
   if (pending)
     playing = (pending == GST_STATE_PLAYING);
   else
@@ -462,7 +684,6 @@ set_progress (ClutterGstVideoTexture *video_texture,
               gdouble                 progress)
 {
   ClutterGstVideoTexturePrivate *priv = video_texture->priv;
-  GstState pending;
   GstQuery *duration_q;
   gint64 position;
 
@@ -471,22 +692,26 @@ set_progress (ClutterGstVideoTexture *video_texture,
 
   CLUTTER_GST_NOTE (MEDIA, "set progress: %.02f", progress);
 
+  priv->in_eos = FALSE;
   priv->target_progress = progress;
 
-  if (priv->in_seek)
+  if (priv->in_download_buffering)
     {
-      CLUTTER_GST_NOTE (MEDIA, "already seeking. stacking progress point.");
+      /* we clear the virtual_stream_buffer_signalled flag as it's likely we
+       * need to buffer again */
+      priv->virtual_stream_buffer_signalled = 0;
+    }
+
+  if (priv->in_seek || priv->is_idle || priv->is_changing_uri)
+    {
+      /* We can't seek right now, let's save the position where we
+         want to seek and do that later. */
+      CLUTTER_GST_NOTE (MEDIA,
+                        "already seeking/idleing. stacking progress point.");
       priv->stacked_progress = progress;
       return;
     }
 
-  gst_element_get_state (priv->pipeline, &priv->stacked_state, &pending, 0);
-
-  if (pending)
-    priv->stacked_state = pending;
-
-  gst_element_set_state (priv->pipeline, GST_STATE_PAUSED);
-
   duration_q = gst_query_new_duration (GST_FORMAT_TIME);
 
   if (gst_element_query (priv->pipeline, duration_q))
@@ -512,6 +737,8 @@ set_progress (ClutterGstVideoTexture *video_texture,
 
   priv->in_seek = TRUE;
   priv->stacked_progress = 0.0;
+
+  CLUTTER_GST_NOTE (MEDIA, "set progress (seeked): %.02f", progress);
 }
 
 static gdouble
@@ -524,12 +751,28 @@ get_progress (ClutterGstVideoTexture *video_texture)
   if (!priv->pipeline)
     return 0.0;
 
+  /* when hitting an error or after an EOS, playbin2 has some weird values when
+   * querying the duration and progress. We default to 0.0 on error and 1.0 on
+   * EOS */
+  if (priv->in_error)
+    {
+      CLUTTER_GST_NOTE (MEDIA, "get progress (error): 0.0");
+      return 0.0;
+    }
+
+  if (priv->in_eos)
+    {
+      CLUTTER_GST_NOTE (MEDIA, "get progress (eos): 1.0");
+      return 1.0;
+    }
+
   /* When seeking, the progress returned by playbin2 is 0.0. We want that to be
    * the last known position instead as returning 0.0 will have some ugly
    * effects, say on a progress bar getting updated from the progress tick. */
-  if (priv->in_seek)
+  if (priv->in_seek || priv->is_changing_uri)
     {
-      CLUTTER_GST_NOTE (MEDIA, "get progress: %.02f", priv->target_progress);
+      CLUTTER_GST_NOTE (MEDIA, "get progress (target): %.02f",
+                        priv->target_progress);
       return priv->target_progress;
     }
 
@@ -550,11 +793,11 @@ get_progress (ClutterGstVideoTexture *video_texture)
     }
   else
     progress = 0.0;
-  
+
   gst_query_unref (position_q);
   gst_query_unref (duration_q);
 
-  CLUTTER_GST_NOTE (MEDIA, "get progress: %.02f", progress);
+  CLUTTER_GST_NOTE (MEDIA, "get progress (pipeline): %.02f", progress);
 
   return progress;
 }
@@ -600,8 +843,10 @@ set_audio_volume (ClutterGstVideoTexture *video_texture,
 
   CLUTTER_GST_NOTE (MEDIA, "set volume: %.02f", volume);
 
-  /* the :volume property is in the [0, 10] interval */
-  g_object_set (G_OBJECT (priv->pipeline), "volume", volume * 10.0, NULL);
+  volume = CLAMP (volume, 0.0, 1.0);
+  gst_stream_volume_set_volume (GST_STREAM_VOLUME (priv->pipeline),
+				GST_STREAM_VOLUME_FORMAT_CUBIC,
+				volume);
   g_object_notify (G_OBJECT (video_texture), "audio-volume");
 }
 
@@ -609,19 +854,13 @@ static gdouble
 get_audio_volume (ClutterGstVideoTexture *video_texture)
 {
   ClutterGstVideoTexturePrivate *priv = video_texture->priv;
-  gdouble volume = 0.0;
 
   if (!priv->pipeline)
     return 0.0;
 
-  /* the :volume property is in the [0, 10] interval */
-  g_object_get (priv->pipeline, "volume", &volume, NULL);
+  CLUTTER_GST_NOTE (MEDIA, "get volume: %.02f", priv->volume);
 
-  volume = CLAMP (volume / 10.0, 0.0, 1.0);
-
-  CLUTTER_GST_NOTE (MEDIA, "get volume: %.02f", volume);
-
-  return volume;
+  return priv->volume;
 }
 
 static void
@@ -878,27 +1117,23 @@ static void
 clutter_gst_video_texture_dispose (GObject *object)
 {
   ClutterGstVideoTexture        *self;
-  ClutterGstVideoTexturePrivate *priv; 
+  ClutterGstVideoTexturePrivate *priv;
 
-  self = CLUTTER_GST_VIDEO_TEXTURE(object); 
+  self = CLUTTER_GST_VIDEO_TEXTURE(object);
   priv = self->priv;
 
   /* FIXME: flush an errors off bus ? */
   /* gst_bus_set_flushing (priv->bus, TRUE); */
 
-  if (priv->pipeline) 
+  /* start by doing the usual clean up when not wanting to play an URI */
+  set_uri (self, NULL);
+
+  if (priv->pipeline)
     {
-      gst_element_set_state (priv->pipeline, GST_STATE_NULL);
       gst_object_unref (GST_OBJECT (priv->pipeline));
       priv->pipeline = NULL;
     }
 
-  if (priv->tick_timeout_id > 0) 
-    {
-      g_source_remove (priv->tick_timeout_id);
-      priv->tick_timeout_id = 0;
-    }
-
   G_OBJECT_CLASS (clutter_gst_video_texture_parent_class)->dispose (object);
 }
 
@@ -906,7 +1141,7 @@ static void
 clutter_gst_video_texture_finalize (GObject *object)
 {
   ClutterGstVideoTexture        *self;
-  ClutterGstVideoTexturePrivate *priv; 
+  ClutterGstVideoTexturePrivate *priv;
 
   self = CLUTTER_GST_VIDEO_TEXTURE (object);
   priv = self->priv;
@@ -920,9 +1155,9 @@ clutter_gst_video_texture_finalize (GObject *object)
 }
 
 static void
-clutter_gst_video_texture_set_property (GObject      *object, 
+clutter_gst_video_texture_set_property (GObject      *object,
 				        guint         property_id,
-				        const GValue *value, 
+				        const GValue *value,
 				        GParamSpec   *pspec)
 {
   ClutterGstVideoTexture *video_texture = CLUTTER_GST_VIDEO_TEXTURE (object);
@@ -974,9 +1209,9 @@ clutter_gst_video_texture_set_property (GObject      *object,
 }
 
 static void
-clutter_gst_video_texture_get_property (GObject    *object, 
+clutter_gst_video_texture_get_property (GObject    *object,
 				        guint       property_id,
-				        GValue     *value, 
+				        GValue     *value,
 				        GParamSpec *pspec)
 {
   ClutterGstVideoTexture *video_texture;
@@ -1116,6 +1351,25 @@ clutter_gst_video_texture_class_init (ClutterGstVideoTextureClass *klass)
                               CLUTTER_GST_SEEK_FLAG_NONE,
                               CLUTTER_GST_PARAM_READWRITE);
   g_object_class_install_property (object_class, PROP_SEEK_FLAGS, pspec);
+
+  /**
+   * ClutterGstVideoTexture::download-buffering:
+   * @start: Start of the buffer (between 0.0 and 1.0)
+   * @stop: End of the buffer (between 0.0 and 1.0)
+   *
+   * When streaming, GStreamer can cache the data in a buffer on the disk,
+   * something called progressive download or download buffering. This signal
+   * is fired when this streaming mode.
+   */
+  signals[DOWNLOAD_BUFFERING] =
+    g_signal_new ("download-buffering",
+                  G_TYPE_FROM_CLASS (object_class),
+                  G_SIGNAL_RUN_LAST,
+                  G_STRUCT_OFFSET (ClutterGstVideoTextureClass,
+                                   download_buffering),
+                  NULL, NULL,
+                  _clutter_gst_marshal_VOID__DOUBLE_DOUBLE,
+                  G_TYPE_NONE, 2, G_TYPE_DOUBLE, G_TYPE_DOUBLE);
 }
 
 static void
@@ -1126,12 +1380,16 @@ bus_message_error_cb (GstBus                 *bus,
   GError *error = NULL;
   ClutterGstVideoTexturePrivate *priv = video_texture->priv;
 
-  gst_message_parse_error (message, &error, NULL);
-        
-  g_signal_emit_by_name (video_texture, "error", error);
-
   gst_element_set_state(priv->pipeline, GST_STATE_NULL);
 
+  gst_message_parse_error (message, &error, NULL);
+
+  /* restore the idle material so we don't just display the last frame */
+  priv->is_idle = TRUE;
+  clutter_actor_queue_redraw (CLUTTER_ACTOR (video_texture));
+
+  g_signal_emit_by_name (video_texture, "error", error);
+
   g_error_free (error);
 }
 
@@ -1141,17 +1399,19 @@ bus_message_eos_cb (GstBus                 *bus,
                     ClutterGstVideoTexture *video_texture)
 {
   ClutterGstVideoTexturePrivate *priv = video_texture->priv;
-  g_object_notify (G_OBJECT (video_texture), "progress");
 
   CLUTTER_GST_NOTE (MEDIA, "EOS");
 
-  gst_element_set_state(priv->pipeline, GST_STATE_NULL);
+  priv->in_eos = TRUE;
+
+  gst_element_set_state(priv->pipeline, GST_STATE_READY);
 
   /* restore the idle material so we don't just display the last frame */
   priv->is_idle = TRUE;
   clutter_actor_queue_redraw (CLUTTER_ACTOR (video_texture));
 
   g_signal_emit_by_name (video_texture, "eos");
+  g_object_notify (G_OBJECT (video_texture), "progress");
 }
 
 static void
@@ -1160,24 +1420,78 @@ bus_message_buffering_cb (GstBus                 *bus,
                           ClutterGstVideoTexture *video_texture)
 {
   ClutterGstVideoTexturePrivate *priv = video_texture->priv;
-  const GstStructure *str;
+  GstBufferingMode mode;
+  GstState current_state;
   gint buffer_percent;
-  gboolean res;
 
-  str = gst_message_get_structure (message);
-  if (!str)
-    return;
+  gst_message_parse_buffering_stats (message, &mode, NULL, NULL, NULL);
 
-  res = gst_structure_get_int (str, "buffer-percent", &buffer_percent);
-  if (res)
+  if (mode != GST_BUFFERING_DOWNLOAD)
+    priv->in_download_buffering = FALSE;
+
+  switch (mode)
     {
-      priv->buffer_fill = CLAMP ((gdouble) buffer_percent / 100.0,
-                                 0.0,
-                                 1.0);
+    case GST_BUFFERING_STREAM:
+      gst_message_parse_buffering (message, &buffer_percent);
+      priv->buffer_fill = CLAMP ((gdouble) buffer_percent / 100.0, 0.0, 1.0);
 
-      CLUTTER_GST_NOTE (MEDIA, "buffer-fill: %.02f", priv->buffer_fill);
+      CLUTTER_GST_NOTE (BUFFERING, "buffer-fill: %.02f", priv->buffer_fill);
+
+      /* The playbin2 documentation says that we need to pause the pipeline
+       * when there's not enough data yet. We try to limit the calls to
+       * gst_element_set_state() */
+      gst_element_get_state (priv->pipeline, &current_state, NULL, 0);
+
+      if (priv->buffer_fill < 1.0)
+        {
+          if (current_state != GST_STATE_PAUSED)
+            {
+              CLUTTER_GST_NOTE (BUFFERING, "pausing the pipeline");
+              gst_element_set_state (priv->pipeline, GST_STATE_PAUSED);
+            }
+        }
+      else
+        {
+          if (current_state != priv->target_state)
+            {
+              CLUTTER_GST_NOTE (BUFFERING, "restoring the pipeline");
+              gst_element_set_state (priv->pipeline, priv->target_state);
+            }
+        }
 
       g_object_notify (G_OBJECT (video_texture), "buffer-fill");
+      break;
+
+    case GST_BUFFERING_DOWNLOAD:
+      /* we rate limit the messages from GStreamer for a usage in a UI (we
+       * don't want *that* many updates). This is done by installing an idle
+       * handler querying the buffer range and sending a signal from there */
+
+      if (priv->in_download_buffering)
+        break;
+
+      /* install the querying idle handler the first time we receive a download
+       * buffering message */
+      configure_buffering_timeout (video_texture, BUFFERING_TIMEOUT);
+
+      /* pause the stream. the idle timeout will set the target state when
+       * having received enough data. We'll use buffer_fill as a "virtual
+       * stream buffer" to signal the application we're buffering until we
+       * can play back from the downloaded stream. */
+      gst_element_set_state (priv->pipeline, GST_STATE_PAUSED);
+      priv->buffer_fill = 0.0;
+      g_object_notify (G_OBJECT (video_texture), "buffer-fill");
+
+      priv->download_buffering_element = g_object_ref (message->src);
+      priv->in_download_buffering = TRUE;
+      priv->virtual_stream_buffer_signalled = 0;
+      break;
+
+    case GST_BUFFERING_TIMESHIFT:
+    case GST_BUFFERING_LIVE:
+    default:
+      g_warning ("Buffering mode %d not handled", mode);
+      break;
     }
 }
 
@@ -1212,8 +1526,15 @@ bus_message_state_change_cb (GstBus                 *bus,
 
   gst_message_parse_state_changed (message, &old_state, &new_state, NULL);
 
-  if (old_state == GST_STATE_READY && 
-      new_state == GST_STATE_PAUSED) 
+  CLUTTER_GST_NOTE (MEDIA, "state change:  %s -> %s",
+                    gst_state_to_string (old_state),
+                    gst_state_to_string (new_state));
+
+  if (old_state == new_state)
+    return;
+
+  if (old_state == GST_STATE_READY &&
+      new_state == GST_STATE_PAUSED)
     {
       GstQuery *query;
 
@@ -1254,7 +1575,18 @@ bus_message_state_change_cb (GstBus                 *bus,
   if (new_state == GST_STATE_NULL)
     priv->is_idle = TRUE;
   else if (new_state == GST_STATE_PLAYING)
-    priv->is_idle = FALSE;
+    {
+      priv->is_idle = FALSE;
+      priv->is_changing_uri = FALSE;
+    }
+
+  if (!priv->is_idle)
+    {
+      if (priv->stacked_progress)
+        {
+          set_progress (video_texture, priv->stacked_progress);
+        }
+    }
 }
 
 static void
@@ -1269,7 +1601,6 @@ bus_message_async_done_cb (GstBus                 *bus,
       g_object_notify (G_OBJECT (video_texture), "progress");
 
       priv->in_seek = FALSE;
-      gst_element_set_state (priv->pipeline, priv->stacked_state);
 
       if (priv->stacked_progress)
         {
@@ -1286,6 +1617,34 @@ on_source_changed (GstElement             *pipeline,
   set_user_agent (video_texture, video_texture->priv->user_agent);
 }
 
+static gboolean
+on_volume_changed_main_context (gpointer data)
+{
+  ClutterGstVideoTexture *video_texture = CLUTTER_GST_VIDEO_TEXTURE (data);
+  ClutterGstVideoTexturePrivate *priv = video_texture->priv;
+  gdouble volume;
+
+  volume = gst_stream_volume_get_volume (GST_STREAM_VOLUME (priv->pipeline),
+					 GST_STREAM_VOLUME_FORMAT_CUBIC);
+  priv->volume = volume;
+
+  g_object_notify (G_OBJECT (video_texture), "audio-volume");
+
+  return FALSE;
+}
+
+/* playbin2 proxies the volume property change notification directly from
+ * the element having the "volume" property. This means this callback is
+ * called from the thread that runs the element, potentially different from
+ * the main thread */
+static void
+on_volume_changed (GstElement             *pipeline,
+		   GParamSpec             *pspec,
+		   ClutterGstVideoTexture *video_texture)
+{
+  g_idle_add (on_volume_changed_main_context, video_texture);
+}
+
 static gboolean
 lay_pipeline (ClutterGstVideoTexture *video_texture)
 {
@@ -1294,7 +1653,7 @@ lay_pipeline (ClutterGstVideoTexture *video_texture)
   GstElement *video_sink = NULL;
 
   priv->pipeline = gst_element_factory_make ("playbin2", "pipeline");
-  if (!priv->pipeline) 
+  if (!priv->pipeline)
     {
       g_critical ("Unable to create playbin2 element");
       return FALSE;
@@ -1308,10 +1667,10 @@ lay_pipeline (ClutterGstVideoTexture *video_texture)
    * FIXME - there must be a way to ask gstreamer to do this for us
    */
   audio_sink = gst_element_factory_make ("gconfaudiosink", "audio-sink");
-  if (!audio_sink) 
+  if (!audio_sink)
     {
       audio_sink = gst_element_factory_make ("autoaudiosink", "audio-sink");
-      if (!audio_sink) 
+      if (!audio_sink)
 	{
 	  audio_sink = gst_element_factory_make ("alsasink", "audio-sink");
 	  g_warning ("Could not create a GST audio_sink. "
@@ -1355,9 +1714,14 @@ clutter_gst_video_texture_init (ClutterGstVideoTexture *video_texture)
 
   priv->is_idle = TRUE;
   priv->in_seek = FALSE;
+  priv->is_changing_uri = FALSE;
+  priv->in_download_buffering = FALSE;
 
   priv->par_n = priv->par_d = 1;
 
+  /* We default to not playing until someone calls set_playing(TRUE) */
+  priv->target_state = GST_STATE_PAUSED;
+
   /* Default to a fast seek, ie. same effect than set_seek_flags (NONE); */
   priv->seek_flags = GST_SEEK_FLAG_KEY_UNIT;
 
@@ -1384,6 +1748,9 @@ clutter_gst_video_texture_init (ClutterGstVideoTexture *video_texture)
                            G_CALLBACK (bus_message_async_done_cb),
                            video_texture, 0);
 
+  g_signal_connect (priv->pipeline, "notify::volume",
+		    G_CALLBACK (on_volume_changed), video_texture);
+
   gst_object_unref (GST_OBJECT (bus));
 }
 
@@ -1421,7 +1788,7 @@ ClutterActor*
 clutter_gst_video_texture_new (void)
 {
   return g_object_new (CLUTTER_GST_TYPE_VIDEO_TEXTURE,
-                       "disable-slicing", TRUE, 
+                       "disable-slicing", TRUE,
                        NULL);
 }
 
@@ -1621,3 +1988,63 @@ clutter_gst_video_texture_set_seek_flags (ClutterGstVideoTexture *texture,
   else if (flags & CLUTTER_GST_SEEK_FLAG_ACCURATE)
     priv->seek_flags = GST_SEEK_FLAG_ACCURATE;
 }
+
+/**
+ * clutter_gst_video_texture_get_buffering_mode:
+ * @texture: a #ClutterGstVideoTexture
+ *
+ * Return value: a #ClutterGstBufferingMode
+ *
+ * Since: 1.4
+ */
+ClutterGstBufferingMode
+clutter_gst_video_texture_get_buffering_mode (ClutterGstVideoTexture *texture)
+{
+  ClutterGstVideoTexturePrivate *priv;
+  GstPlayFlags flags;
+
+  g_return_val_if_fail (CLUTTER_GST_IS_VIDEO_TEXTURE (texture),
+                        CLUTTER_GST_BUFFERING_MODE_STREAM);
+  priv = texture->priv;
+
+  g_object_get (G_OBJECT (priv->pipeline), "flags", &flags, NULL);
+  if (flags & GST_PLAY_FLAG_DOWNLOAD)
+    return CLUTTER_GST_BUFFERING_MODE_DOWNLOAD;
+
+  return CLUTTER_GST_BUFFERING_MODE_STREAM;
+}
+
+/**
+ * clutter_gst_video_texture_set_buffering_mode:
+ * @texture: a #ClutterGstVideoTexture
+ * @mode: a #ClutterGstBufferingMode
+ *
+ * Since: 1.4
+ */
+void
+clutter_gst_video_texture_set_buffering_mode (ClutterGstVideoTexture *texture,
+                                              ClutterGstBufferingMode mode)
+{
+  ClutterGstVideoTexturePrivate *priv;
+  GstPlayFlags flags;
+
+  g_return_if_fail (CLUTTER_GST_IS_VIDEO_TEXTURE (texture));
+  priv = texture->priv;
+
+  g_object_get (G_OBJECT (priv->pipeline), "flags", &flags, NULL);
+
+  switch (mode)
+    {
+    case CLUTTER_GST_BUFFERING_MODE_STREAM:
+      flags &= ~GST_PLAY_FLAG_DOWNLOAD;
+      break;
+    case CLUTTER_GST_BUFFERING_MODE_DOWNLOAD:
+      flags |= GST_PLAY_FLAG_DOWNLOAD;
+      break;
+    default:
+      g_warning ("Unexpected buffering mode %d", mode);
+      break;
+    }
+
+  g_object_set (G_OBJECT (priv->pipeline), "flags", flags, NULL);
+}
