@@ -25,6 +25,7 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <stdlib.h>
 #include <string.h>
 
 #include <glib/gprintf.h>
@@ -134,7 +135,6 @@ main (int argc, char *argv[])
    * efficient/corrent playback onto the texture (which sucks a bit)
    */
   texture = g_object_new (CLUTTER_TYPE_TEXTURE,
-                          "sync-size",       FALSE,
                           "disable-slicing", TRUE,
                           NULL);
 
